///////////////////////////////////////////////////
// File:	PhysicsMath.cpp
// Author:	Edward D. Baum
// Purpose:	Math structures and functions required
//			for PhysicsObject.
//
///////////////////////////////////////////////////

#include <stdio.h>

#include "PhysicsMath.h"

const Vector kVector0(0.0, 0.0, 0.0, 0.0);
const Vector kPoint0(0.0, 0.0, 0.0, 1.0);
const Vector kVectorX(1.0, 0.0, 0.0, 0.0);
const Vector kVectorY(0.0, 1.0, 0.0, 0.0);
const Vector kVectorZ(0.0, 0.0, 1.0, 0.0);

const Matrix kMatrix1(Vector(1.0, 0.0, 0.0, 0.0),
					  Vector(0.0, 1.0, 0.0, 0.0),
					  Vector(0.0, 0.0, 1.0, 0.0),
					  Vector(0.0, 0.0, 0.0, 1.0));

int Vector::mWrite(char *iBuffer) const
{
	return sprintf(iBuffer, "%f\t%f\t%f\t%f", fX, fY, fZ, fW);
}

int Matrix::mWrite(char *iBuffer) const
{
	return sprintf(iBuffer, "%f\t%f\t%f\t%f\n%f\t%f\t%f\t%f\n%f\t%f\t%f\t%f\n%f\t%f\t%f\t%f",
		fRight.fX, fFront.fX, fUp.fX, fPosition.fX,
		fRight.fY, fFront.fY, fUp.fY, fPosition.fY,
		fRight.fZ, fFront.fZ, fUp.fZ, fPosition.fZ,
		fRight.fW, fFront.fW, fUp.fW, fPosition.fW);
}

bool gSpherePlaneIntersect(
	const Vector &iSphereStart, const Vector &iSphereEnd, const double iSphereRadius,
	const Vector &iPlanePoint, const Vector &iPlaneNormal,
	Vector &iCollisionPoint, Vector &iCollisionNormal, double &iCollisionTime)
{
	// Pretend plane goes through (0, 0, 0)
	Vector lSphereStart = iSphereStart;
	lSphereStart -= iPlanePoint;
	double lStartDist = lSphereStart * iPlaneNormal;

	// First check if we start colliding, as this could lead to
	//  negative front or back plane collision
	if(lStartDist <= iSphereRadius && lStartDist >= 0.0)
	{
		// We start out coliding in front of the plane
		// The collision point is the point directly toward the plane,
		//  the start distance away.
		iCollisionPoint = iSphereStart;
		iCollisionPoint -= iPlaneNormal * lStartDist;
		// The collision Normal is the plane normal
		iCollisionNormal = iPlaneNormal;
		// The Collision Time is 0, it starts colliding
		iCollisionTime = 0.0;

		return true;
	}
	else if(lStartDist >= -iSphereRadius && lStartDist <= 0.0)
	{
		// We start out coliding in back of the plane
		// The collision point is the point directly toward the plane,
		//  the start distance away.
		iCollisionPoint = iSphereStart;
		iCollisionPoint += iPlaneNormal * lStartDist;
		// The collision Normal is the plane normal negated
		iCollisionNormal = -iPlaneNormal;
		// The Collision Time is 0, it starts colliding
		iCollisionTime = 0.0;

		return true;
	}

	// We don't normalize this as we need it's length for proper time calculation.
	Vector lSphereDirection = iSphereEnd;
	lSphereDirection -= iSphereStart;

	// Check that we are not parralel, and that we are moving towards the plane
	double lDenominator = lSphereDirection * iPlaneNormal;
	double lCollisionTime;
	if(lDenominator * lStartDist >= 0.0)
		return false;

	// We will test one of two paralel planes which are iSphereRadius away.
	//  We test the front plane if we start in front, and the back plane if
	//  we start in back.
	if(lStartDist >= 0.0)
	{
		// Test front Plane.  Move start point and keep
		//  pretending plane goes through (0, 0, 0).
		//  Just say that we started closer.
		lStartDist -= iSphereRadius;
		lCollisionTime = -lStartDist / lDenominator;
		if(lCollisionTime <= 1.0) // We would have caught < 0.0 already, with moving towards check
		{
			// The collision point is the point we are at at
			//  the time we collide.
			iCollisionPoint = iSphereStart;
			iCollisionPoint += lSphereDirection * lCollisionTime;
			iCollisionPoint -= iPlaneNormal * iSphereRadius;
			// The collision Normal is the plane normal
			iCollisionNormal = iPlaneNormal;
			// The Collision Time is...the collision time
			iCollisionTime = lCollisionTime;

			return true;
		}
	}
	else
	{
		// Test front Plane.  Move start point and keep
		//  pretending plane goes through (0, 0, 0).
		//  Just say that we started closer.
		lStartDist += iSphereRadius;
		lCollisionTime = -lStartDist / lDenominator;
		if(lCollisionTime <= 1.0) // We would have caught < 0.0 already, with moving towards check
		{
			// The collision point is the point we are at at
			//  the time we collide.
			iCollisionPoint = iSphereStart;
			iCollisionPoint += lSphereDirection * lCollisionTime;
			iCollisionPoint += iPlaneNormal * iSphereRadius;
			// The collision Normal is the plane normal negated
			iCollisionNormal = -iPlaneNormal;
			// The Collision Time is...the collision time
			iCollisionTime = lCollisionTime;

			return true;
		}
	}

	return false;
}
