///////////////////////////////////////////////////
// File:	PhysicsMath.h
// Author:	Edward D. Baum
// Purpose:	Math structures and functions required
//			for PhysicsObject.
//
// Notes:	No assignment operators return any
//			value.  This is done for the marginal
//			increase in efficiency.
///////////////////////////////////////////////////

#ifndef _PHYSICS_MATH_H
#define _PHYSICS_MATH_H

#include <math.h>

static const double kPI = 3.14159265358979323846;

inline double gAbsoluteValue(double iNumber)
	{return ((iNumber < 0.0) ? -iNumber : iNumber);};

inline double gMax(double iLHS, double iRHS)
	{return ((iLHS > iRHS) ? iLHS : iRHS);};

inline double gMin(double iLHS, double iRHS)
	{return ((iLHS < iRHS) ? iLHS : iRHS);};

class Vector
{
public:
	// Constructors
	inline Vector() {};
	inline Vector(double iX, double iY, double iZ, double iW = 0.0)
		{fX = iX; fY = iY; fZ = iZ; fW = iW;};

	// Scaler Operations
	inline void operator*=(double iNumber)
		{fX *= iNumber; fY *= iNumber; fZ *= iNumber; fW *= iNumber;};

	inline void operator/=(double iNumber) // Not safe
		{fX /= iNumber; fY /= iNumber; fZ /= iNumber; fW /= iNumber;};

	inline Vector operator*(double iNumber) const
		{Vector iToReturn; iToReturn.fX = fX * iNumber; iToReturn.fY = fY * iNumber;
		iToReturn.fZ = fZ * iNumber; iToReturn.fW = fW * iNumber; return iToReturn;};

	inline Vector operator/(double iNumber) const // Not safe
		{Vector iToReturn; iToReturn.fX = fX / iNumber; iToReturn.fY = fY / iNumber;
		iToReturn.fZ = fZ / iNumber; iToReturn.fW = fW / iNumber; return iToReturn;};

	inline friend Vector operator*(double iNumber, const Vector &iVector)
		{return iVector * iNumber;};

	// Vector Operations
	inline void operator=(const Vector &iRHS)
		{fX = iRHS.fX; fY = iRHS.fY; fZ = iRHS.fZ; fW = iRHS.fW;}

	inline void operator+=(const Vector &iRHS)
		{fX += iRHS.fX; fY += iRHS.fY; fZ += iRHS.fZ; fW += iRHS.fW;};

	inline void operator-=(const Vector &iRHS)
		{fX -= iRHS.fX; fY -= iRHS.fY; fZ -= iRHS.fZ; fW -= iRHS.fW;};

	inline friend Vector operator+(const Vector &iLHS, const Vector &iRHS)
		{Vector iToReturn; iToReturn.fX = iLHS.fX + iRHS.fX; iToReturn.fY = iLHS.fY + iRHS.fY;
		iToReturn.fZ = iLHS.fZ + iRHS.fZ; iToReturn.fW = iLHS.fW + iRHS.fW; return iToReturn;};

	inline friend Vector operator-(const Vector &iLHS, const Vector &iRHS)
		{Vector iToReturn; iToReturn.fX = iLHS.fX - iRHS.fX; iToReturn.fY = iLHS.fY - iRHS.fY;
		iToReturn.fZ = iLHS.fZ - iRHS.fZ; iToReturn.fW = iLHS.fW - iRHS.fW; return iToReturn;};

	inline Vector operator-() const
		{Vector lToReturn; lToReturn.fX = -fX; lToReturn.fY = -fY;
		lToReturn.fZ = -fZ; lToReturn.fW = -fW; return lToReturn;};

	inline double operator*(const Vector &iRHS) const
		{return ((fX * iRHS.fX) + (fY * iRHS.fY) + (fZ * iRHS.fZ) + (fW * iRHS.fW));};

	inline Vector operator%(const Vector &iRHS) const // Three dimensional cross product
		{Vector iToReturn; iToReturn.fX = (fY * iRHS.fZ) - (fZ * iRHS.fY);
		iToReturn.fY = (fZ * iRHS.fX) - (fX * iRHS.fZ);
		iToReturn.fZ = (fX * iRHS.fY) - (fY * iRHS.fX);
		iToReturn.fW = 0.0; return iToReturn;};

	inline void mCrossProduct(const Vector &iLHS, const Vector &iRHS) // Three dimensional
		{fX = (iLHS.fY * iRHS.fZ) - (iLHS.fZ * iRHS.fY);
		fY = (iLHS.fZ * iRHS.fX) - (iLHS.fX * iRHS.fZ);
		fZ = (iLHS.fX * iRHS.fY) - (iLHS.fY * iRHS.fX);
		fW = 0.0;};

	inline double mGetLength() const
		{return sqrt((*this) * (*this));};

	inline void mNormalize()
		{(*this) /= this->mGetLength();};

	// Output Functions
	int mWrite(char *iBuffer) const; // Buffer length is caller's responsibility.

	// Data
	double fX;
	double fY;
	double fZ;
	double fW;
};

extern const Vector kVector0;
extern const Vector kPoint0;
extern const Vector kVectorX;
extern const Vector kVectorY;
extern const Vector kVectorZ;

// Just to be clear if we right matrices the way mathmaticians do this matrix is in the form:
//
// [ fRight.fX fFront.fX fUp.fX fPosition.fX]
// [ fRight.fY fFront.fY fUp.fY fPosition.fY]
// [ fRight.fZ fFront.fZ fUp.fX fPosition.fZ]
// [ fRight.fW fFront.fW fUp.fW fPosition.fW]
//
//  and in this form M * v will take a point in the matrix space and
//  put it in the wider world space.
class Matrix
{
public:
	// Constructors
	inline Matrix() {};
	inline Matrix(const Vector &iRight, const Vector &iFront,
		const Vector &iUp, const Vector &iPosition)
		{fRight = iRight; fFront = iFront; fUp = iUp; fPosition = iPosition;};

	// Scaler Functions
	inline void operator*=(double iNumber)
		{fRight *= iNumber; fFront *= iNumber; fUp *= iNumber; fPosition *= iNumber;};

	inline void operator/=(double iNumber) // Not safe
		{fRight /= iNumber; fFront /= iNumber; fUp /= iNumber; fPosition /= iNumber;};

	inline Matrix operator*(double iNumber) const
		{Matrix iToReturn; iToReturn.fRight = fRight * iNumber; iToReturn.fFront = fFront * iNumber;
		iToReturn.fUp = fUp * iNumber; iToReturn.fPosition = fPosition * iNumber; return iToReturn;};

	inline Matrix operator/(double iNumber) const // Not safe
		{Matrix iToReturn; iToReturn.fRight = fRight / iNumber; iToReturn.fFront = fFront / iNumber;
		iToReturn.fUp = fUp / iNumber; iToReturn.fPosition = fPosition / iNumber; return iToReturn;};

	inline friend Matrix operator*(double iNumber, const Matrix &iMatrix)
		{return iMatrix * iNumber;};

	// Vector Functions
	inline Vector operator*(const Vector &iRHS) const
		{
			Vector lToReturn;
			lToReturn.fX = fRight.fX * iRHS.fX + fFront.fX * iRHS.fY
				+ fUp.fX * iRHS.fZ + fPosition.fX * iRHS.fW;
			lToReturn.fY = fRight.fY * iRHS.fX + fFront.fY * iRHS.fY
				+ fUp.fY * iRHS.fZ + fPosition.fY * iRHS.fW;
			lToReturn.fZ = fRight.fZ * iRHS.fX + fFront.fZ * iRHS.fY
				+ fUp.fZ * iRHS.fZ + fPosition.fZ * iRHS.fW;
			lToReturn.fW = fRight.fW * iRHS.fX + fFront.fW * iRHS.fY
				+ fUp.fW * iRHS.fZ + fPosition.fW * iRHS.fW;
			return lToReturn;
		};

	inline void mMultiply(const Vector &iOperand, Vector &iProduct) const
		{
			iProduct.fX = fRight.fX * iOperand.fX + fFront.fX * iOperand.fY
				+ fUp.fX * iOperand.fZ + fPosition.fX * iOperand.fW;
			iProduct.fY = fRight.fY * iOperand.fX + fFront.fY * iOperand.fY
				+ fUp.fY * iOperand.fZ + fPosition.fY * iOperand.fW;
			iProduct.fZ = fRight.fZ * iOperand.fX + fFront.fZ * iOperand.fY
				+ fUp.fZ * iOperand.fZ + fPosition.fZ * iOperand.fW;
			iProduct.fW = fRight.fW * iOperand.fX + fFront.fW * iOperand.fY
				+ fUp.fW * iOperand.fZ + fPosition.fW * iOperand.fW;
		};

	// Matrix Functions
	inline void operator=(const Matrix &iMatrix)
		{fRight = iMatrix.fRight; fFront = iMatrix.fFront;
		fUp = iMatrix.fUp; fPosition = iMatrix.fPosition;};

	inline void operator+=(const Matrix &iRHS)
		{fRight += iRHS.fRight; fFront += iRHS.fFront;
		fUp += iRHS.fUp; fPosition += iRHS.fPosition;};

	inline void operator-=(const Matrix &iRHS)
		{fRight -= iRHS.fRight; fFront -= iRHS.fFront;
		fUp -= iRHS.fUp; fPosition -= iRHS.fPosition;};

	inline void operator*=(const Matrix &iRHS)
		{
			Vector lTemp;

			lTemp.fX = fRight.fX; lTemp.fY = fFront.fX; lTemp.fZ = fUp.fX; lTemp.fW = fPosition.fX;
			fRight.fX = lTemp * iRHS.fRight; fFront.fX = lTemp * iRHS.fFront;
			fUp.fX = lTemp * iRHS.fUp; fPosition.fX = lTemp * iRHS.fPosition;

			lTemp.fX = fRight.fY; lTemp.fY = fFront.fY; lTemp.fZ = fUp.fY; lTemp.fW = fPosition.fY;
			fRight.fY = lTemp * iRHS.fRight; fFront.fY = lTemp * iRHS.fFront;
			fUp.fY = lTemp * iRHS.fUp; fPosition.fY = lTemp * iRHS.fPosition;

			lTemp.fX = fRight.fZ; lTemp.fY = fFront.fZ; lTemp.fZ = fUp.fZ; lTemp.fW = fPosition.fZ;
			fRight.fZ = lTemp * iRHS.fRight; fFront.fZ = lTemp * iRHS.fFront;
			fUp.fZ = lTemp * iRHS.fUp; fPosition.fZ = lTemp * iRHS.fPosition;

			lTemp.fX = fRight.fW; lTemp.fY = fFront.fW; lTemp.fZ = fUp.fW; lTemp.fW = fPosition.fW;
			fRight.fW = lTemp * iRHS.fRight; fFront.fW = lTemp * iRHS.fFront;
			fUp.fW = lTemp * iRHS.fUp; fPosition.fW = lTemp * iRHS.fPosition;
		};

	inline friend Matrix operator*(const Matrix &iLHS, const Matrix &iRHS) // A little slower than optimal, maybe
		{
			Matrix lToReturn;

			Vector lTemp;

			lTemp.fX = iLHS.fRight.fX; lTemp.fY = iLHS.fFront.fX;
			lTemp.fZ = iLHS.fUp.fX; lTemp.fW = iLHS.fPosition.fX;
			lToReturn.fRight.fX = lTemp * iRHS.fRight;lToReturn.fFront.fX = lTemp * iRHS.fFront;
			lToReturn.fUp.fX = lTemp * iRHS.fUp; lToReturn.fPosition.fX = lTemp * iRHS.fPosition;

			lTemp.fX = iLHS.fRight.fY; lTemp.fY = iLHS.fFront.fY;
			lTemp.fZ = iLHS.fUp.fY; lTemp.fW = iLHS.fPosition.fY;
			lToReturn.fRight.fY = lTemp * iRHS.fRight; lToReturn.fFront.fY = lTemp * iRHS.fFront;
			lToReturn.fUp.fY = lTemp * iRHS.fUp; lToReturn.fPosition.fY = lTemp * iRHS.fPosition;

			lTemp.fX = iLHS.fRight.fZ; lTemp.fY = iLHS.fFront.fZ;
			lTemp.fZ = iLHS.fUp.fZ; lTemp.fW = iLHS.fPosition.fZ;
			lToReturn.fRight.fZ = lTemp * iRHS.fRight; lToReturn.fFront.fZ = lTemp * iRHS.fFront;
			lToReturn.fUp.fZ = lTemp * iRHS.fUp; lToReturn.fPosition.fZ = lTemp * iRHS.fPosition;

			lTemp.fX = iLHS.fRight.fW; lTemp.fY = iLHS.fFront.fW;
			lTemp.fZ = iLHS.fUp.fW; lTemp.fW = iLHS.fPosition.fW;
			lToReturn.fRight.fW = lTemp * iRHS.fRight; lToReturn.fFront.fW = lTemp * iRHS.fFront;
			lToReturn.fUp.fW = lTemp * iRHS.fUp; lToReturn.fPosition.fW = lTemp * iRHS.fPosition;

			return lToReturn;
		};

	inline void mGetTransposition(Matrix &iTransposition)
		{
			iTransposition.fRight.fX = fRight.fX; iTransposition.fRight.fY = fFront.fX;
			iTransposition.fRight.fZ = fUp.fX; iTransposition.fRight.fW = fPosition.fX;

			iTransposition.fFront.fX = fRight.fY; iTransposition.fFront.fY = fFront.fY;
			iTransposition.fFront.fZ = fUp.fY; iTransposition.fFront.fW = fPosition.fY;

			iTransposition.fUp.fX = fRight.fZ; iTransposition.fUp.fY = fFront.fZ;
			iTransposition.fUp.fZ = fUp.fZ; iTransposition.fUp.fW = fPosition.fZ;

			iTransposition.fPosition.fX = fRight.fW; iTransposition.fPosition.fY = fFront.fW;
			iTransposition.fPosition.fZ = fUp.fW; iTransposition.fPosition.fW = fPosition.fW;
		}

	inline void mGetInverse(Matrix &iInverse) const // 3D inverse
		{
			double lDeterminant =	fRight.fX * (fFront.fY * fUp.fZ - fFront.fZ * fUp.fY)
									- fFront.fX * (fRight.fY * fUp.fZ - fRight.fZ * fUp.fY)
									+ fUp.fX * (fRight.fY * fFront.fZ - fRight.fZ * fFront.fY);

			if(!lDeterminant)
			{
				iInverse.fRight = kVectorX;
				iInverse.fFront = kVectorY;
				iInverse.fUp = kVectorZ;
				iInverse.fPosition = kPoint0;

				return;
			}

			iInverse.fRight.fX = (fFront.fY * fUp.fZ - fFront.fZ * fUp.fY) / lDeterminant;
			iInverse.fRight.fY = (fUp.fY * fRight.fZ - fRight.fY * fUp.fZ) / lDeterminant;
			iInverse.fRight.fZ = (fRight.fY * fFront.fZ - fRight.fZ * fFront.fY) / lDeterminant;
			iInverse.fRight.fW = 0.0;

			iInverse.fFront.fX = (fUp.fX * fFront.fZ - fFront.fX * fUp.fZ) / lDeterminant;
			iInverse.fFront.fY = (fRight.fX * fUp.fZ - fUp.fX * fRight.fZ) / lDeterminant;
			iInverse.fFront.fZ = (fFront.fX * fRight.fZ - fRight.fX * fFront.fZ) / lDeterminant;
			iInverse.fFront.fW = 0.0;

			iInverse.fUp.fX = (fFront.fX * fUp.fY - fUp.fX * fFront.fY) / lDeterminant;
			iInverse.fUp.fY = (fUp.fX * fRight.fY - fRight.fX * fUp.fY) / lDeterminant;
			iInverse.fUp.fZ = (fRight.fX * fFront.fY - fFront.fX * fRight.fY) / lDeterminant;
			iInverse.fUp.fW = 0.0;

			iInverse.fPosition = kPoint0;
		}

	inline void mSkewSymetric(const Vector &iMakeFrom)
		{
			fRight.fX = 0.0;
			fRight.fY = iMakeFrom.fZ;
			fRight.fZ = -iMakeFrom.fY;
			fRight.fW = 0.0;
			fFront.fX = -iMakeFrom.fZ;
			fFront.fY = 0.0;
			fFront.fZ = iMakeFrom.fX;
			fFront.fW = 0.0;
			fUp.fX = iMakeFrom.fY;
			fUp.fY = -iMakeFrom.fX;
			fUp.fZ = 0.0;
			fUp.fW = 0.0;
			fPosition = kVector0;
		};

	inline void mProjection(float iHalfAngleOfViewY, float iAspectRatio, float iNear, float iFar)
	{
		float lTan = tan(iHalfAngleOfViewY);
		float lHalfHeight = lTan * iNear;
		float lHalfWidth = lHalfHeight * iAspectRatio;		
		
		fRight.fX = iNear / lHalfWidth;
		fRight.fY = 0.0;
		fRight.fZ = 0.0;
		fRight.fW = 0.0;
		fFront.fX = 0.0;
		fFront.fY = iNear / lHalfHeight;
		fFront.fZ = 0.0;
		fFront.fW = 0.0;
		fUp.fX = 0.0;
		fUp.fY = 0.0;
		fUp.fZ = (iFar + iNear) / (iNear - iFar);
		fUp.fW = -1;
		fPosition.fX = 0.0;
		fPosition.fY = 0.0;
		fPosition.fZ = 2.0 * iNear * iFar / (iNear - iFar);
		fPosition.fW = 0.0;
	} 

	inline void mCamera(const Vector &iFrom, const Vector& iTo, const Vector &iUp) {
		// First find the camera space to world space transformation.
		fFront = iTo;
		fFront -= iFrom;
		fFront.mNormalize();
		fRight.mCrossProduct(fFront, iUp);
		fRight.mNormalize();
		fUp.mCrossProduct(fRight, fFront); // Already normalized as both others are length 1.
		fUp.mNormalize();
		// Don't bother, we're about to overwrite.
		//fPosition = iFrom;

		// Now we invert the camera to world space transformation.
		// We take advantage of the fact the the 3x3 rotation
		//  sub-matrix is orthogonal already.  This means
		//  the inverse is just the transpose.  The position is
		//  the inverse of the camera position (the world space origin
		//  in (non-rotated) camera space) transformed by the camera to world
		//  rotation.  This is just multiplication by our pre-transposed
		//  vectors, conveniently.
		// Do this first while it's easy.
		fPosition.fX = -(fRight * iFrom);
		fPosition.fY = -(fFront * iFrom);
		fPosition.fZ = -(fUp * iFrom);

		float lTemp = fRight.fY;
		fRight.fY = fFront.fX;
		fFront.fX = lTemp;
		lTemp = fRight.fZ;
		fRight.fZ = fUp.fX;
		fUp.fX = lTemp;
		lTemp = fFront.fZ;
		fFront.fZ = fUp.fY;
		fUp.fY = lTemp;

		fRight.fW = 0.0;
		fFront.fW = 0.0;
		fUp.fW = 0.0;
		fPosition.fW = 1.0;

		// We do a little more here because we actually want X to be right,
		//  Y to be up, and -Z to be forward while we currently have X right,
		//  Y forward, and Z Up.  This is like multiplying the point by
		//  a transformation matrix before we multiply by the camera to
		//  world matrix we currently have.
		//  If we were in an object space which was rotated this way then
		//  to get to world space we would use the matrix:
		//  [1  0  0  0]
		//  [0  0  1  0]
		//  [0 -1  0  0]
		//  [0  0  0  1]
		//  We want to invert this.  As above we can just transpose to:
		//  [1  0  0  0]
		//  [0  0 -1  0]
		//  [0  1  0  0]
		//  [0  0  0  1]
		// This is, not surprisingly, equivalent to swapping the second and
		//  third rows of the matrix and making the new third row negative.
		Vector lTempV;
		lTempV.fX = -fRight.fY; lTempV.fY = -fFront.fY; lTempV.fZ = -fUp.fY; lTempV.fW = -fPosition.fY;
		fRight.fY = fRight.fZ; fFront.fY = fFront.fZ; fUp.fY = fUp.fZ; fPosition.fY = fPosition.fZ;
		fRight.fZ = lTempV.fX; fFront.fZ = lTempV.fY; fUp.fZ = lTempV.fZ; fPosition.fZ = lTempV.fW;
	}

	inline void mOrthoNormalize() // Three dimensional
		{
			fFront.mNormalize();
			fRight.mCrossProduct(fFront, fUp);
			fRight.mNormalize();
			fUp.mCrossProduct(fRight, fFront);
			fUp.mNormalize();
		}

	// Output Functions
	int mWrite(char *iBuffer) const; // Buffer length is caller's responsibility.

	// Data
	Vector fRight;
	Vector fFront;
	Vector fUp;
	Vector fPosition;
};

extern const Matrix kMatrix1;

// Plane doesn't move, sphere does
bool gSpherePlaneIntersect(
	const Vector &iSphereStart, const Vector &iSphereEnd, const double iSphereRadius,
	const Vector &iPlanePoint, const Vector &iPlaneNormal,
	Vector &iCollisionPoint, Vector &iCollisionNormal, double &iCollisionTime);

#endif // _PHYSICS_MATH_H
