///////////////////////////////////////////////////
// File:	PhysicsObject.cpp
// Author:	Edward D. Baum
// Purpose:	A rigid body object for simulation.
//
///////////////////////////////////////////////////

#include "PhysicsObject.h"

const double DYNAMIC_LIMIT = 0.00001;
const double ROLLING_LIMIT = 0.00001;
const double SPIN_CLOSENESS_LIMIT = 0.999;

Vector PhysicsObject::sGravity(0.0, 0.0, -9.8, 0.0);

PhysicsObject::PhysicsObject()
{
	Vector lVector0(0.0, 0.0, 0.0, 0.0);
	Matrix lMatrix1(Vector(1.0, 0.0, 0.0, 0.0),
					Vector(0.0, 1.0, 0.0, 0.0),
					Vector(0.0, 0.0, 1.0, 0.0),
					Vector(0.0, 0.0, 0.0, 1.0));

	fOrientation				= lMatrix1;
	fLastOrientation			= lMatrix1;
	fVelocity					= lVector0;
	fAngularVelocity			= lVector0;
	fAngularMomentum			= lVector0;
	fForce						= lVector0;
	fTorque						= lVector0;
	fPartCount					= 0;
	fMassInverse				= 1.0;
	fBaseMomentOfInertiaInverse	= lMatrix1;
	fMomentOfInertiaInverse		= lMatrix1;
	fCoeficientOfRestitution	= 0.5;
	fDynamicFrictionCoeficient	= 0.5;
	fRollingFrictionCoeficient	= 0.2;
	fSpinningFrictionCoeficient	= 0.2;
	fDragCoeficient				= 100.0;
	fFrameTime					= 0.0;
	fDensity					= 1.0;
}

void PhysicsObject::mInteropolateValues(double iTime)
{
	Vector lPosition = fOrientation.fPosition;
	fOrientation.fPosition = kVector0;
	Matrix lOrientationTranspose;
	fVelocity += (iTime * fMassInverse) * fForce;
	fAngularMomentum += iTime * fTorque;

	fMomentOfInertiaInverse = fOrientation;
	fMomentOfInertiaInverse *= fBaseMomentOfInertiaInverse;
	fOrientation.mGetTransposition(lOrientationTranspose);
	fMomentOfInertiaInverse *= lOrientationTranspose;

	fOrientation.fPosition = lPosition;

	fMomentOfInertiaInverse.mMultiply(fAngularMomentum, fAngularVelocity);
}

void PhysicsObject::mStartPhysicsFrame(double iTime)
{
	fFrameTime = 0.0;

	mPredict(iTime);
}

void PhysicsObject::mEndPhysicsFrame(double iTime)
{
	fOrientation = fEndOrientation;

	mInteropolateValues(iTime);
}

void PhysicsObject::mStartGeneralFrame()
{
	fLastOrientation = fOrientation;
}

void PhysicsObject::mEndGeneralFrame()
{
	fLastOrientation = fOrientation;
	fForce = PhysicsObject::sGravity;
	fForce /= fMassInverse;
	fTorque = kVector0;
}

void PhysicsObject::mPredict(double iTime)
{
	double lTime = iTime - fFrameTime;
	Matrix lAngularVelocityMatrix;

	fEndOrientation = fOrientation;

	Vector lEndOrientationPosition = fEndOrientation.fPosition + lTime * fVelocity;
	fEndOrientation.fPosition = kVector0;
	lAngularVelocityMatrix.mSkewSymetric(fAngularVelocity);
	lAngularVelocityMatrix *= fEndOrientation;
	lAngularVelocityMatrix *= lTime;
	fEndOrientation += lAngularVelocityMatrix;
	fEndOrientation.fPosition = lEndOrientationPosition;

	fEndOrientation.mOrthoNormalize();
}

void PhysicsObject::mMove(double iTime)
{
	double lTime = iTime - fFrameTime;
	fFrameTime = iTime;
	Matrix lAngularVelocityMatrix;

	Vector lOrientationPosition = fOrientation.fPosition + lTime * fVelocity;
	fOrientation.fPosition = kVector0;
	lAngularVelocityMatrix.mSkewSymetric(fAngularVelocity);
	lAngularVelocityMatrix *= fOrientation;
	lAngularVelocityMatrix *= lTime;
	fOrientation += lAngularVelocityMatrix;
	fOrientation.fPosition = lOrientationPosition;

	fOrientation.mOrthoNormalize();
}

// TODO: I'm working on the simpler stuff first
bool PhysicsObject::mCollide(Vector &iCollisionSpot, Vector &iCollisionNormal,
	double &iCollisionTime, const PhysicsObject &iToTest, double iStartTime, double iEndTime) const
{
	return false;
}

bool PhysicsObject::mTerrainColide(const Vector &iPlanePoint, const Vector &iPlaneNormal,
	Vector &iCollisionSpot, Vector &iCollisionNormal,
	double &iCollisionTime, double iStartTime, double iEndTime) const
{
	// Avoid some copies if possible
	Matrix lStartBuffer;
	//Matrix lEndBuffer;
	// If the manager is working correctly iStartTime >= fFrameTime
	const Matrix *lStartMatrix = (iStartTime > fFrameTime) ? &lStartBuffer : &fOrientation;
	// For now the end time is always the end of the physics frame
	const Matrix *lEndMatrix = &fEndOrientation;

	// Need to keep potential collisions
	Vector lCollidingPoint;
	Vector lCollidingNormal;
	double lCollidingTime = 2.0;	// Bigger so that collisions at
									//  the end time can be detected.

	// To use as reference arguments to find each collision
	Vector lCollisionPoint;
	Vector lCollisionNormal;
	double lCollisionTime;

	// Need to know each spheres position at start and end
	Vector lSphereStart;
	Vector lSphereEnd;

	// Need to know where the ground is
	Vector lGroundPoint = kPoint0;
	Vector lGroundNormal = kVectorZ;

	// If the manager is working correctly iStartTime >= fFrameTime
	if(iStartTime > fFrameTime)
	{
		double lTimeToStart = iStartTime - fFrameTime;
		Matrix lAngularVelocityMatrix;
		lStartBuffer = fOrientation;
		Vector lStartBufferPosition = lStartBuffer.fPosition + lTimeToStart * fVelocity;
		lStartBuffer.fPosition = kVector0;
		lAngularVelocityMatrix.mSkewSymetric(fAngularVelocity);
		lAngularVelocityMatrix *= fOrientation;
		lAngularVelocityMatrix *= lTimeToStart;
		lStartBuffer += lAngularVelocityMatrix;
		lStartBuffer.fPosition = lStartBufferPosition;
		lStartBuffer.mOrthoNormalize();
	}

	// Test to see for each spere if it has gone through the ground
	//  If it has we record this penetration if it is before our
	//  current first
	for(int i = 0; i < fPartCount; ++i)
	{
		Vector lSpherePos = fParts[i];
		lSpherePos.fW = 1.0; // As we encode the radius here.
		lStartMatrix->mMultiply(lSpherePos, lSphereStart);
		lEndMatrix->mMultiply(lSpherePos, lSphereEnd);

		// If collide with the ground
		if(gSpherePlaneIntersect(lSphereStart, lSphereEnd, fParts[i].fW,
			iPlanePoint, iPlaneNormal, lCollisionPoint, lCollisionNormal, lCollisionTime)
			&& (lCollisionTime < lCollidingTime))
		{
				lCollidingPoint = lCollisionPoint;
				lCollidingNormal = lCollisionNormal;
				lCollidingTime = lCollisionTime;
		}
	}

	if(lCollidingTime <= 1.0)
	{
		iCollisionSpot = lCollidingPoint;
		iCollisionNormal = lCollidingNormal;
		iCollisionTime = iStartTime + lCollidingTime * (iEndTime - iStartTime);
		return true;
	}
	else
		return false;
}

void PhysicsObject::mResolveCollision(const Vector &iCollisionSpot,
	const Vector &iCollisionNormal, PhysicsObject &iHitObject, double iEndTime)
{
	return;
}

void PhysicsObject::mResolveTerrainCollision(const Vector &iCollisionSpot,
	const Vector &iCollisionNormal, double iEndTime)
{
	Vector lLocalSpot = iCollisionSpot - fOrientation.fPosition;
	Vector lLocalVelocity = fVelocity + fAngularVelocity % lLocalSpot;
	Vector lNormalVelocity = (lLocalVelocity * iCollisionNormal) * iCollisionNormal;
	Vector lParallelVelocity = lLocalVelocity - lNormalVelocity;
	double lParallelMagnitude = lParallelVelocity * lParallelVelocity;
	double lCollisionMagnitude = mTerrainImpulseMagnitude(iCollisionNormal, lLocalSpot,
		lLocalVelocity) * (1.0 + fCoeficientOfRestitution);

	mApplyImpulse(lCollisionMagnitude * iCollisionNormal, iCollisionSpot);

	// Friction
	if(lParallelMagnitude > DYNAMIC_LIMIT)
	{
		// Dynamic
		lParallelMagnitude  = sqrt(lParallelMagnitude);
		lParallelVelocity /= lParallelMagnitude;
		double lDynamicFrictionMagnitude =
			gMin(lCollisionMagnitude * fDynamicFrictionCoeficient, lParallelMagnitude / fMassInverse);
		mApplyImpulse(-lDynamicFrictionMagnitude * lParallelVelocity, iCollisionSpot);
	}

	// Rolling
	Vector lVelocity = fVelocity - (fVelocity * iCollisionNormal) * iCollisionNormal;
	double lVelocityMagnitude = sqrt(lVelocity * lVelocity);
	if(lVelocityMagnitude > ROLLING_LIMIT)
	{
		double lRollingFrictionMagnitude =
			gMin(lCollisionMagnitude * fRollingFrictionCoeficient / sqrt(lLocalSpot * lLocalSpot), lVelocityMagnitude / fMassInverse);
		mApplyImpulse(-lRollingFrictionMagnitude * (lVelocity / lVelocityMagnitude));
	}

	// Spinning.  Only apply when spinning on the collision point.
	lLocalSpot.mNormalize(); // We can do this as we do not use it again
	if((lLocalSpot * iCollisionNormal) < SPIN_CLOSENESS_LIMIT)
	{
		double lSpinComponent =  fAngularMomentum * iCollisionNormal;
		double lSpinningFrictionMagnitude ;
		if(lSpinComponent >= 0.0)
			lSpinningFrictionMagnitude =
				gMin(lCollisionMagnitude * fSpinningFrictionCoeficient, lSpinComponent);
		else
			lSpinningFrictionMagnitude =
				gMax(-lCollisionMagnitude * fSpinningFrictionCoeficient, lSpinComponent);
		fAngularMomentum -= lSpinningFrictionMagnitude * iCollisionNormal;
	}

	mPredict(iEndTime);
}

Vector PhysicsObject::mCollisionImpulse(const Vector &iCollisionSpot,
	const Vector &iCollisionNormal, const PhysicsObject &iHitObject) const
{
	return kVector0;
}

Vector PhysicsObject::mTerrainCollisionImpulse(const Vector &iCollisionSpot,
	const Vector &iCollisionNormal) const
{
	Vector lLocalSpot = iCollisionSpot - fOrientation.fPosition;
	Vector lLocalVelocity = fVelocity + lLocalSpot % fAngularVelocity;

	return  mTerrainImpulseMagnitude(iCollisionNormal, lLocalSpot, lLocalVelocity) *
		(1.0 + fCoeficientOfRestitution) * iCollisionNormal;
}

void PhysicsObject::mApplyForce(const Vector &iForce)
{
	fForce += iForce;
}

void PhysicsObject::mApplyForce(const Vector &iForce, const Vector &iPosition)
{
	fForce += iForce;
	fTorque += (iPosition - fOrientation.fPosition) % iForce;
}

void PhysicsObject::mApplyTorque(const Vector &iTorque)
{
	fTorque += iTorque;
}

void PhysicsObject::mApplyImpulse(const Vector &iImpulse)
{
	fVelocity += iImpulse * fMassInverse;
}

void PhysicsObject::mApplyImpulse(const Vector &iImpulse, const Vector &iPosition)
{
	fVelocity += iImpulse * fMassInverse;
	fAngularMomentum += (iPosition - fOrientation.fPosition) % iImpulse;
	fMomentOfInertiaInverse.mMultiply(fAngularMomentum, fAngularVelocity);
}

void PhysicsObject::mSetInertia(const Matrix &iMomentOfInertia)
{
	Vector lPosition = fOrientation.fPosition;
	fOrientation.fPosition = kVector0;
	Matrix lOrientationTranspose;
	iMomentOfInertia.mGetInverse(fBaseMomentOfInertiaInverse);
	fOrientation.mGetTransposition(lOrientationTranspose);

	fMomentOfInertiaInverse = fOrientation;
	fMomentOfInertiaInverse *= fBaseMomentOfInertiaInverse;
	fMomentOfInertiaInverse *= lOrientationTranspose;

	fOrientation.fPosition = lPosition;
}
