///////////////////////////////////////////////////
// File:	PhysicsObject.h
// Author:	Edward D. Baum
// Purpose:	A rigid body object for simulation.
//
///////////////////////////////////////////////////

#ifndef _PHYSICS_OBJECT_H
#define _PHYSICS_OBJECT_H

#include "PhysicsMath.h"

const int MAX_OBJECT_PARTS = 16;

class PhysicsObject
{
public:
	// Just a constructor
	PhysicsObject();

	// Update our quantities at the end of a simulation frame.
	//  Apply forces, torques, update moment of inertia, etc.
	//  This should be the last thing called at the end of a
	//  physics frame in order to prepare for the next one.
	// It is called within mEndPhysicsFrame
	void mInteropolateValues(double iTime);

	// Prepare for a round of physics
	void mStartPhysicsFrame(double iTime);

	// Close off this physics frame for this object
	void mEndPhysicsFrame(double iTime);

	// Store previous orientation mainly
	void mStartGeneralFrame();

	// Clear accumulators for force, torque, etc. for the
	//  rest of the simulation
	void mEndGeneralFrame();

	// Set where we expect to be at the end of a frame.
	//  Frame (physics) must end to move us there.
	void mPredict(double iTime);

	// Move.  Sets fFrameTime. Should be called only by collision
	void mMove(double iTime);

	// Do collision.  See if our move actualy worked out.  Test against one object.
	bool mCollide(Vector &iCollisionSpot, Vector &iCollisionNormal, double &iCollisionTime,
		const PhysicsObject &iToTest, double iStartTime, double iEndTime) const;
	bool mTerrainColide(const Vector &iPlanePoint, const Vector &iPlaneNormal,
		Vector &iCollisionSpot, Vector &iCollisionNormal,
		double &iCollisionTime, double iStartTime, double iEndTime) const;

	// Resolve a collision if you don't feel like doing extra messy work.  Requires that
	//  the object already be moved.  Useful because it will allow contact forces and friction.
	void mResolveCollision(const Vector &iCollisionSpot, const Vector &iCollisionNormal,
		PhysicsObject &iHitObject, double iEndTime);
	void mResolveTerrainCollision(const Vector &iCollisionSpot,
		const Vector &iCollisionNormal, double iEndTime);

	// Find impulse after a collision (to apply to the object which collided),
	//  you must move object to correct time first.
	Vector mCollisionImpulse(const Vector &iCollisionSpot, const Vector &iCollisionNormal,
		const PhysicsObject &iHitObject) const;
	Vector mTerrainCollisionImpulse(const Vector &iCollisionSpot,
		const Vector &iCollisionNormal) const;

	// Apply a force until the end of a frame (game).
	//  Considered to start at start of current physics
	//  frame or start of next game frame.
	void mApplyForce(const Vector &iForce);
	void mApplyForce(const Vector &iForce, const Vector &iPosition);

	// Apply a torque until the end of a frame (game).
	//  Considered to start at start of current physics
	//  frame or start of next game frame.
	void mApplyTorque(const Vector &iTorque);

	// Apply an impulse immediatly
	// Only effective if mPredict called at the time Impulse is applied
	void mApplyImpulse(const Vector &iImpulse);
	void mApplyImpulse(const Vector &iImpulse, const Vector &iPosition);

	// For the physics manager
	inline double mGetFrameTime() const
		{return fFrameTime;};

	// Set various things.  Dangerous to call in the middle of a physics frame.
	inline void mSetOrientation(const Matrix &iOrientation)
		{fOrientation = iOrientation;};
	inline void mSetVelocity(const Vector &iVelocity)
		{fVelocity = iVelocity;}
	inline void mSetAngularMomentum(const Vector &iAngularMomentum)
		{fAngularMomentum = iAngularMomentum;};
	inline void mAddPart(const Vector &iPart)
		{if(fPartCount < MAX_OBJECT_PARTS) {fParts[fPartCount++] = iPart;}};
	inline void mSetMass(double iMass)
		{fMassInverse = 1.0f / iMass;};
	void mSetInertia(const Matrix &iMomentOfInertia);
	inline void mSetRestitution(double iCoefficient)
		{fCoeficientOfRestitution = gMin(gMax(iCoefficient, 0.0f), 1.0f);};
	inline void mSetDynamicFriction(double iDynamicFrictionCoeficient)
		{fDynamicFrictionCoeficient = iDynamicFrictionCoeficient;};
	inline void mSetRollingFriction(double iRollingFrictionCoeficient)
		{fRollingFrictionCoeficient = iRollingFrictionCoeficient;};
	inline void mSetSpinningFriction(double iSpinningFrictionCoeficient)
		{fSpinningFrictionCoeficient = iSpinningFrictionCoeficient;};

	// Get all sorts of values
	inline Matrix& mGetOrientation()
		{return fOrientation;};
	inline Matrix& mGetLastOrientation()
		{return fLastOrientation;};
	inline Vector& mGetVelocity()
		{return fVelocity;};
	inline Vector& mGetAngularMomentum()
		{return fAngularMomentum;};
	inline double mGetMass() const
		{return 1.0f / fMassInverse;};
	inline double mGetRestitutionCoeficient() const
		{return fCoeficientOfRestitution;};
	inline double mGetDynamicFriction() const
		{return fDynamicFrictionCoeficient;};
	inline double mGetRollingFriction() const
		{return fRollingFrictionCoeficient;};
	inline double mGetSpinningFriction() const
		{return fSpinningFrictionCoeficient;};


	// Static functions
	inline static void sSetGravity(const Vector &iGravity)
		{sGravity = iGravity;};
	inline static Vector sGetGravity()
		{return sGravity;};

protected:

	inline double mTerrainImpulseMagnitude(const Vector &iCollisionNormal,
		const Vector &iLocalSpot, const Vector &iLocalVelocity) const
	{
		Vector lInertiaDamping;
		fMomentOfInertiaInverse.mMultiply(iLocalSpot % iCollisionNormal, lInertiaDamping);

		return (gAbsoluteValue(iLocalVelocity * iCollisionNormal) /
			(((iCollisionNormal * iCollisionNormal) * fMassInverse) +
			((lInertiaDamping % iLocalSpot) * iCollisionNormal)));
	};

	// Various forces
	inline void mApplyDrag()
	{
		mApplyForce((-fDragCoeficient * fVelocity.mGetLength()) * fVelocity);
	}

	Matrix fOrientation;
	Vector fVelocity;
	Matrix fEndOrientation; // Our expected orientation at the end of a physics frame
	Matrix fLastOrientation; // At start of general frame
	Vector fAngularVelocity;
	Vector fAngularMomentum;
	Vector fForce;
	Vector fTorque;
	Vector fParts[MAX_OBJECT_PARTS];
	int fPartCount;
	double fMassInverse;
	Matrix fMomentOfInertiaInverse;
	Matrix fBaseMomentOfInertiaInverse;
	double fCoeficientOfRestitution;
	double fDynamicFrictionCoeficient;
	double fRollingFrictionCoeficient;
	double fSpinningFrictionCoeficient;
	double fDragCoeficient;
	double fFrameTime; // For multiple collisions, how far we are into the frame.
	double fDensity;

	static Vector sGravity;
};

#endif // _PHYSICS_OBJECT_H
