///////////////////////////////////////////////////
// File:	Main.cpp
// Author:	Edward D. Baum
// Purpose:	A simple test of the physics package.
//
///////////////////////////////////////////////////

#ifdef EMSCRIPTEN
#include <emscripten.h>
#endif //EMSCRIPTEN

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <SDL/SDL.h>

#include <GLES2/gl2platform.h>
#include <GLES2/gl2.h>

#include "PhysicsObject.h"

static const int kPhysicsTimeMultiple = 10;

static const double kHalfAngleOfView = kPI * 0.125; // 22.5 degrees
static const double kNearClipDistance = 0.1;
static const double kFarClipDistance = 10000.0;

static bool gActive = true;

static Vector gLookAt(0., -9., 0., 1.);
static Vector gLookFrom(0., -10., 0., 1.);
static Vector gLookOffset(0., 1., 0., 0.);

static PhysicsObject gBall;

static Vector gSides[6];
static Vector gSideNormals[6];
static Vector gSideCorners[6];
static Vector gSideAdjacentCorners[6];
static double gSideAspectRatios[6];
static double gFieldHalfWidth = 30.0;
static double gFieldHalfHeight = 5.0;
static double gMaxLineSeperation = 1.0;

static double gPitch = 0;
static double gYaw = 0;

static char gHorizontalControl = 0;
static char gVerticalControl = 0;
static char gMoveControl = 0;
static char gPushControl = 0;

static float gPushForce = 20.0;

static const double kHorizontalSpeed = kPI / 4;	// radians / s
static const double kVerticalSpeed = kPI / 4;	// radians / s
static const double kCameraSpeed = 5.0;			// units / s
static bool gFollowMode = false;

static int gTime = 0;
static bool gPaused = false;

SDL_Surface *gpSurface = NULL;

static GLuint gVertexShader = 0;
static GLuint gFragmentShader = 0;
static GLuint gShaderProgram = 0;
static const char *kVertexShaderSource =
"precision mediump float;\n"
"attribute vec3 vPosition;\n"
"uniform mat4 mProjection;\n"
"uniform mat4 mCamera;\n"
"uniform mat4 mObject;\n"
"\n"
"void main(void) {\n"
"	gl_Position = mProjection * mCamera * mObject * vec4(vPosition, 1.0);\n"
"}";
static const char *kFragmentShaderSource =
"precision mediump float;\n"
"uniform vec4 cColor;\n"
"\n"
"void main(void) {\n"
"	gl_FragColor = cColor;\n"
"}";

static GLfloat gObjectMat[16];
static GLfloat gCameraMat[16];
static GLfloat gProjMat[16];

// Some helper classes.
struct KeyNode
{
	KeyNode(unsigned int iKey) : fKey(iKey), fNext(0) {}
	
	unsigned int fKey;
	KeyNode *fNext;
};

class KeyList
{
	public:
		KeyList() : fHead(NULL) {}
		~KeyList()
		{
			KeyNode *lCurrent = fHead;
			while(lCurrent)
			{
				KeyNode *lNext = lCurrent->fNext;
				delete lCurrent;
				lCurrent = lNext;
			}
		}

		bool mAdd(unsigned int iKey)
		{
			KeyNode **fLast = &fHead;
			KeyNode *lCurrent = fHead;
			while(lCurrent) {
				if(lCurrent->fKey == iKey)
				 return false;

				fLast = &lCurrent->fNext;
				lCurrent = lCurrent->fNext;
			}

			*fLast = new KeyNode(iKey);
			return true;
		}

		bool mRemove(unsigned int iKey)
		{
			KeyNode **fLast = &fHead;
			KeyNode *lCurrent = fHead;
			while(lCurrent) {
				if(lCurrent->fKey == iKey)
				{
					*fLast = lCurrent->fNext;
					delete lCurrent;
					
					return true;
				}
				 

				fLast = &lCurrent->fNext;
				lCurrent = lCurrent->fNext;
			}
		
			return false;
		}

		bool mCheck(unsigned int iKey)
		{
			for(KeyNode *lCurrent = fHead; lCurrent; lCurrent = lCurrent->fNext)
			{
				if(lCurrent->fKey == iKey)
					return true;
			}
		
			return false;
		}

	protected:
		KeyNode *fHead;
} gKeysDown;

GLuint gCreateShader(GLenum iShaderType, const char *iShaderSource)
{
	GLint lSourceLength = strlen(iShaderSource);
	GLuint lShaderID = glCreateShader(iShaderType);
	glShaderSource(lShaderID, 1, &iShaderSource, &lSourceLength);
	glCompileShader(lShaderID);

	GLint lStatus;
	glGetShaderiv(lShaderID, GL_COMPILE_STATUS, &lStatus);
	if(lStatus == GL_FALSE)
	{
		fprintf(stderr, "Oh noes! Shader status 0x%08x.\nSource:\n%s\n", lStatus, iShaderSource);
		GLsizei lCharBufferLength = 512;
		char lCharBuffer[512];
		glGetShaderInfoLog(lShaderID, 512, &lCharBufferLength, lCharBuffer);
		fprintf(stderr, "Program Log (%i): %s\n", lCharBufferLength, lCharBufferLength ? lCharBuffer : "<none>");
		exit(1);
	}

	return lShaderID; 
}

GLuint gCreateProgram(GLuint iVertexShader, GLuint iFragmentShader)
{
	GLuint lShaderProgram = glCreateProgram();
	glAttachShader(lShaderProgram, iVertexShader);
	glAttachShader(lShaderProgram, iFragmentShader);
	glLinkProgram(lShaderProgram);

	GLint lStatus;
	glGetProgramiv(lShaderProgram, GL_LINK_STATUS, &lStatus);
	if(lStatus == GL_FALSE)
	{
		fprintf(stderr, "Oh noes! Program status 0x%08x.\n", lStatus);
		GLsizei lCharBufferLength = 512;
		char lCharBuffer[512];
		glGetProgramInfoLog(lShaderProgram, 512, &lCharBufferLength, lCharBuffer);
		fprintf(stderr, "Program Log (%i): %s\n", lCharBufferLength, lCharBufferLength ? lCharBuffer : "<none>");
		exit(1);
	}

	return lShaderProgram;
}

void gGetGLMatrix(GLfloat *iGLMatrix, const Matrix &iMatrix)
{
	iGLMatrix[0] = iMatrix.fRight.fX;
	iGLMatrix[1] = iMatrix.fRight.fY;
	iGLMatrix[2] = iMatrix.fRight.fZ;
	iGLMatrix[3] = iMatrix.fRight.fW;
	iGLMatrix[4] = iMatrix.fFront.fX;
	iGLMatrix[5] = iMatrix.fFront.fY;
	iGLMatrix[6] = iMatrix.fFront.fZ;
	iGLMatrix[7] = iMatrix.fFront.fW;
	iGLMatrix[8] = iMatrix.fUp.fX;
	iGLMatrix[9] = iMatrix.fUp.fY;
	iGLMatrix[10] = iMatrix.fUp.fZ;
	iGLMatrix[11] = iMatrix.fUp.fW;
	iGLMatrix[12] = iMatrix.fPosition.fX;
	iGLMatrix[13] = iMatrix.fPosition.fY;
	iGLMatrix[14] = iMatrix.fPosition.fZ;
	iGLMatrix[15] = iMatrix.fPosition.fW;
}

class DrawPlane
{
	public:
	DrawPlane() : fLineBuffer(0), fLineBufferID(0), fPoints(0) {}
	DrawPlane(const Vector &iCenter, const Vector &iCorner, const Vector &iAdjacentCorner)
	{
		mInit(iCenter, iCorner, iAdjacentCorner);
	}
	~DrawPlane()
	{
		mShutdown();
	}

	void mInit(const Vector &iCenter, const Vector &iCorner, const Vector &iAdjacentCorner)
	{
		mShutdown();

		Vector lCornerDirection;
		Vector lAdjacentCornerDirection;
	
		Vector lHorizontal;
		Vector lVertical;
		double lWidth;
		double lHeight;
	
		Vector lLineStart;
		Vector lLineEnd;
		Vector lSeperation;
	
		lCornerDirection = iCorner - iCenter;
		lAdjacentCornerDirection = iAdjacentCorner - iCenter;

		Vector lNormal;
		lNormal.mCrossProduct(lCornerDirection, lAdjacentCornerDirection);
		lNormal.mNormalize();
		fNormal[0] = 0.5 * (abs(lNormal.fX) + 1);
		fNormal[1] = 0.5 * (abs(lNormal.fY) + 1);
		fNormal[2] = 0.5 * (abs(lNormal.fZ) + 1);
		fNormal[3] = 1.0;
	
		lHorizontal = lAdjacentCornerDirection;
		lHorizontal -= lCornerDirection;
		lWidth = lHorizontal.mGetLength();
	
		lVertical = lAdjacentCornerDirection;
		lVertical += lCornerDirection;
		lVertical *= -1.0;
		lHeight = lVertical.mGetLength();
	
		int lLinesH = (int)(lWidth / gMaxLineSeperation) + 2; // The +2 is for each end
		int lLinesV = (int)(lHeight / gMaxLineSeperation) + 2; // The +2 is for each end
		int lLines = lLinesH ;
		fPoints = (lLinesH + lLinesV) * 2 * 2; // 2 points per line, horizontal and vertical lines.
		lLineStart = iCorner;
		lLineEnd = lVertical;
		lLineEnd += iCorner;
		lSeperation = lHorizontal;
		lSeperation /= (lLines - 1);
	
		fLineBuffer = new GLfloat[3 * fPoints];
		GLfloat *plCurrent = fLineBuffer;

		for(int i = 0; i < lLines; ++i)
		{
			*plCurrent = lLineStart.fX;
			++plCurrent;
			*plCurrent = lLineStart.fY;
			++plCurrent;
			*plCurrent = lLineStart.fZ;
			++plCurrent;
			*plCurrent = lLineEnd.fX;
			++plCurrent;
			*plCurrent = lLineEnd.fY;
			++plCurrent;
			*plCurrent = lLineEnd.fZ;
			++plCurrent;
			lLineStart += lSeperation;
			lLineEnd += lSeperation;
		}

		lLines = lLinesV;
		fPoints += lLines;
		lLineStart = iCorner;
		lLineEnd = lHorizontal;
		lLineEnd += iCorner;
		lSeperation = lVertical;
		lSeperation /= (lLines - 1);
	
		for(int j = 0; j < lLines; ++j)
		{
			*plCurrent = lLineStart.fX;
			++plCurrent;
			*plCurrent = lLineStart.fY;
			++plCurrent;
			*plCurrent = lLineStart.fZ;
			++plCurrent;
			*plCurrent = lLineEnd.fX;
			++plCurrent;
			*plCurrent = lLineEnd.fY;
			++plCurrent;
			*plCurrent = lLineEnd.fZ;
			++plCurrent;
			lLineStart += lSeperation;
			lLineEnd += lSeperation;
		}

		// Now we do the OpenGL stuff.
		glGenBuffers(1, &fLineBufferID);
		glBindBuffer(GL_ARRAY_BUFFER, fLineBufferID);
		glBufferData(GL_ARRAY_BUFFER, 3 * fPoints * sizeof(GLfloat), fLineBuffer, GL_STATIC_DRAW);
	}

	// Just use the currently set position, camera, and perspective.
	void mDraw()
	{
		glUniform4fv(glGetUniformLocation(gShaderProgram, "cColor"), 1, fNormal);
		glBindBuffer(GL_ARRAY_BUFFER, fLineBufferID);
		glEnableVertexAttribArray(glGetAttribLocation(gShaderProgram, "vPosition"));
		glVertexAttribPointer(glGetAttribLocation(gShaderProgram, "vPosition"), 3, GL_FLOAT, GL_FALSE, 0, 0);
		glDrawArrays(GL_LINES, 0, fPoints);
	}

	protected:
	void mShutdown()
	{
		if(fLineBuffer)
			delete [] fLineBuffer;
		fLineBuffer = NULL;
	}

	GLfloat *fLineBuffer;
	GLfloat fNormal[4];
	GLuint fLineBufferID;
	GLuint fPoints;
} gPlanes[6];

class DrawBall
{
	public:
		DrawBall() : fLineBuffer(NULL), fLineBufferID(0), fPoints(0) {}
		DrawBall(unsigned int iSlices, unsigned int iSectors, float iRadius)
		{
			mInit(iSlices, iSectors, iRadius);
		}
		~DrawBall()
		{
			mShutdown();
		}
		
		
		void mInit(unsigned int iSlices, unsigned int iSectors, float iRadius)
		{
			mShutdown();
			
			typedef Vector* Vector_ptr;
			Vector **lpMeridians = new Vector_ptr[iSectors];
			for(unsigned int i = 0; i < iSectors; ++i)
				lpMeridians[i] = new Vector[iSlices];

			Vector lTop(0.0, 0.0, iRadius, 1.0);
			Vector lBottom(0.0, 0.0, -iRadius, 1.0);

			float lPhiDelta = kPI / iSlices;
			float lThetaDelta = 2.0 * kPI / iSectors;
			float lPhi = lPhiDelta; // Start at the top.
			for(unsigned int i = 0; i < iSlices; ++i)
			{
				float lSliceRadius = iRadius * sin(lPhi);
				float lSliceHeight = iRadius * cos(lPhi);
				float lTheta = 0.0f;
				for(unsigned int j = 0; j < iSectors; ++j)
				{	
					lpMeridians[j][i].fX = lSliceRadius * cos(lTheta);
					lpMeridians[j][i].fY = lSliceRadius * sin(lTheta);
					lpMeridians[j][i].fZ = lSliceHeight;
					lpMeridians[j][i].fW = 1.0;
					
					lTheta += lThetaDelta;
				}
			
				lPhi += lPhiDelta;
			}

			// Create the meridians then create the great circles.
			// TODO: Do this more efficiently.
			fPoints = 2 * ((1 + iSlices) * iSectors + iSlices * iSectors);
			fLineBuffer = new GLfloat[3 * fPoints];
			GLfloat *lpElement = fLineBuffer;

			// Meridians
			for(unsigned int i = 0; i < iSectors; ++i)
			{
				*lpElement = lTop.fX;
				++lpElement;
				*lpElement = lTop.fY;
				++lpElement;
				*lpElement = lTop.fZ;
				++lpElement;
				
				for(unsigned int j = 0; j < iSlices; ++j)
				{
					// Each point is both an end and a start.
					for(unsigned int k = 0; k < 2; ++k)
					{
						*lpElement = lpMeridians[i][j].fX;
						++lpElement;
						*lpElement = lpMeridians[i][j].fY;
						++lpElement;
						*lpElement = lpMeridians[i][j].fZ;
						++lpElement;
					}
				}
			
				*lpElement = lBottom.fX;
				++lpElement;
				*lpElement = lBottom.fY;
				++lpElement;
				*lpElement = lBottom.fZ;
				++lpElement;
			}
		
			// Great Circles.
			for(unsigned int i = 0; i < iSlices; ++i)
			{
				*lpElement = lpMeridians[0][i].fX;
				++lpElement;
				*lpElement = lpMeridians[0][i].fY;
				++lpElement;
				*lpElement = lpMeridians[0][i].fZ;
				++lpElement;
				
				for(unsigned int j = 1; j < iSectors; ++j)
				{
					// Every point is both an end and a start.
					for(unsigned int k = 0; k < 2; ++k)
					{
						*lpElement = lpMeridians[j][i].fX;
						++lpElement;
						*lpElement = lpMeridians[j][i].fY;
						++lpElement;
						*lpElement = lpMeridians[j][i].fZ;
						++lpElement;
					}
				}

				*lpElement = lpMeridians[0][i].fX;
				++lpElement;
				*lpElement = lpMeridians[0][i].fY;
				++lpElement;
				*lpElement = lpMeridians[0][i].fZ;
				++lpElement;
			}
		
			for(unsigned int i = 0; i < iSectors; ++i)
				delete [] lpMeridians[i];
				
			delete [] lpMeridians;
			
		// Now we do the OpenGL stuff.
		glGenBuffers(1, &fLineBufferID);
		glBindBuffer(GL_ARRAY_BUFFER, fLineBufferID);
		glBufferData(GL_ARRAY_BUFFER, 3 * fPoints * sizeof(GLfloat), fLineBuffer, GL_STATIC_DRAW);
	}

	// Just use the currently set position, camera, and perspective.
	void mDraw()
	{
		fColor[0] = fColor[1] = fColor[2] = fColor[3] = 1.0;
		
		glUniform4fv(glGetUniformLocation(gShaderProgram, "cColor"), 1, fColor);
		glBindBuffer(GL_ARRAY_BUFFER, fLineBufferID);
		glEnableVertexAttribArray(glGetAttribLocation(gShaderProgram, "vPosition"));
		glVertexAttribPointer(glGetAttribLocation(gShaderProgram, "vPosition"), 3, GL_FLOAT, GL_FALSE, 0, 0);
		glDrawArrays(GL_LINES, 0, fPoints);
	}

	protected:
		void mShutdown()
		{
			if(fLineBuffer)
				delete [] fLineBuffer;
			fLineBuffer = NULL;
		}
	
		GLfloat *fLineBuffer;
		GLuint fLineBufferID;
		GLuint fPoints;
		GLfloat fColor[4];
} gDrawBall;

void gRender()
{
	// Set up our camera matrix.
	Matrix lCamera;
	lCamera.mCamera(gLookFrom, gLookAt, kVectorZ);
	gGetGLMatrix(gCameraMat, lCamera);

	// Start with not object matrix.
	gGetGLMatrix(gObjectMat, kMatrix1);	
	
	// Clear our buffers
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// Make sure all matrices start in a good state.
	glUniformMatrix4fv(glGetUniformLocation(gShaderProgram, "mObject"), 1, GL_FALSE, gObjectMat);
	glUniformMatrix4fv(glGetUniformLocation(gShaderProgram, "mCamera"), 1, GL_FALSE, gCameraMat);
	glUniformMatrix4fv(glGetUniformLocation(gShaderProgram, "mProjection"), 1, GL_FALSE, gProjMat);

	// Draw Field
	for(int i = 0; i < 6; ++i)
	{
		gPlanes[i].mDraw();
	}

	// TODO: Draw Ball.
	// Change the matrix for the ball.
	gGetGLMatrix(gObjectMat, gBall.mGetOrientation()); 
	glUniformMatrix4fv(glGetUniformLocation(gShaderProgram, "mObject"), 1, GL_FALSE, gObjectMat);
	gDrawBall.mDraw();
	
	glFlush();
	glFinish();
}

void gReshape(int iWidth, int iHeight)
{
	if(!iHeight)
		iHeight = 1;

	double lAspectRatio = ((double)iWidth) / ((double)iHeight);

	// Set our viewport right.
	glViewport(0, 0, iWidth, iHeight);

	// Set up our camera matrix.
	Matrix lCamera;
	lCamera.mProjection(kHalfAngleOfView, lAspectRatio, kNearClipDistance, kFarClipDistance);
	gGetGLMatrix(gProjMat, lCamera);
}

void gProcessKeyDown(unsigned iKey, int iX, int iY)
{
	if(!gKeysDown.mAdd(iKey))
		return;
	
	switch(iKey)
	{
	case ' ':
		gPushControl += 1;
		break;
	case 'a':
	case 'A':
		gMoveControl += 1;
		break;
	case 's':
	case 'S':
		gMoveControl -= 1;
		break;
	case 'p':
	case 'P':
		gPaused = !gPaused;
		break;
	case 'f':
	case 'F':
		gFollowMode = !gFollowMode;
		break;
	case 'g':
	case 'G':
		if(PhysicsObject::sGetGravity().mGetLength())
			PhysicsObject::sSetGravity(kVector0);
		else
			PhysicsObject::sSetGravity(-9.8 * kVectorZ);
		break;
	case 'r':
	case 'R':
		gBall.mSetVelocity(kVector0);
		gBall.mSetOrientation(kMatrix1);
		gBall.mSetAngularMomentum(kVector0);
		break;
	case 'j':
	case 'J':
		{
			const Vector &lVelocity = gBall.mGetVelocity();
			if((lVelocity.fX * lVelocity.fX > 0.0001f) || (lVelocity.fY * lVelocity.fY > 0.0001f))
			{
				gLookOffset = lVelocity;
				gLookOffset.mNormalize();
			} 
			gLookFrom = gBall.mGetOrientation().fPosition;
			gLookAt = gLookFrom + gLookOffset;
			gPitch = asin(gLookOffset.fZ);
			gYaw = acos(gLookOffset.fY /
				sqrt(gLookOffset.fX * gLookOffset.fX + gLookOffset.fY * gLookOffset.fY));
		}
		break;
	case SDLK_UP:
		gVerticalControl += 1;
		break;
	case SDLK_DOWN:
		gVerticalControl -= 1;
		break;
	case SDLK_LEFT:
		gHorizontalControl -= 1;
		break;
	case SDLK_RIGHT:
		gHorizontalControl += 1;
		break;
	default:
		break;
	}
}

void gProcessKeyUp(unsigned iKey, int iX, int iY)
{
	if(!gKeysDown.mRemove(iKey))
		return;
	
	switch(iKey)
	{
	case 'a':
	case 'A':
		gMoveControl -= 1;
		break;
	case 's':
	case 'S':
		gMoveControl += 1;
		break;
	case ' ':
		gPushControl -= 1;
		break;
	default:
		break;
	case SDLK_UP:
		gVerticalControl -= 1;
		break;
	case SDLK_DOWN:
		gVerticalControl += 1;
		break;
	case SDLK_LEFT:
		gHorizontalControl += 1;
		break;
	case SDLK_RIGHT:
		gHorizontalControl -= 1;
		break;
	}
}

void gSimulate()
{
	int lCurrentTime = SDL_GetTicks();
	double lDeltaTime;

	if(gTime < lCurrentTime)
	{
		// Time change
		lDeltaTime = (lCurrentTime - gTime) * 0.001; // Convert to seconds
		if(lDeltaTime > 0.1)
			lDeltaTime = 0.1;
		gTime = lCurrentTime;

		// Ball Simulation
		if(!gPaused)
		{
			// Apply push force to ball, done before frame so that it acts on it.
			if(gPushControl)
     			gBall.mApplyForce(gPushControl * gPushForce * gLookOffset);

			gBall.mStartGeneralFrame();
			double lPhysicsDeltaTime = lDeltaTime / kPhysicsTimeMultiple;
			double lPhysicsFrameTime = 0.0;
			double lCollisionTime;
			Vector lCollisionPoint;
			Vector lCollisionNormal;
			double lFirstCollisionTime = 2.0 * lPhysicsDeltaTime; // So that an actual collision is always less
			Vector lFirstCollisionPoint;
			Vector lFirstCollisionNormal;

			for(int i = 0; i < kPhysicsTimeMultiple; ++i)
			{
				bool lCheckCollision = true;

				gBall.mStartPhysicsFrame(lPhysicsDeltaTime);
				while(lCheckCollision)
				{
					lCheckCollision = false;

					for(int j = 0; j < 6; ++j)
					{
						if(gBall.mTerrainColide(gSides[j], gSideNormals[j],
							lCollisionPoint, lCollisionNormal,
							lCollisionTime, lPhysicsFrameTime, lPhysicsDeltaTime)
							&& ((lCollisionTime > lPhysicsFrameTime)
							|| ((gBall.mGetVelocity() * gSideNormals[j]) < 0))) // One way collision
						{
							if(lCollisionTime < lFirstCollisionTime)
							{
								lFirstCollisionTime = lCollisionTime;
								lFirstCollisionPoint = lCollisionPoint;
								lFirstCollisionNormal = lCollisionNormal;
								lCheckCollision = true;
							}
						}
					}

					if(lCheckCollision)
					{
						gBall.mMove(lFirstCollisionTime);
						gBall.mResolveTerrainCollision(lFirstCollisionPoint, lFirstCollisionNormal, lPhysicsDeltaTime);
						lPhysicsFrameTime = lFirstCollisionTime;
						lFirstCollisionTime = 2.0 * lPhysicsDeltaTime;
					}
				}

				gBall.mEndPhysicsFrame(lPhysicsDeltaTime);
			}

			gBall.mEndGeneralFrame();
		}

		// Camera Change
		if(gMoveControl)
		{
			gLookFrom += gMoveControl * lDeltaTime * kCameraSpeed * gLookOffset;
			gLookAt = gLookFrom + gLookOffset;
		}

		if(gFollowMode)
		{
			gLookOffset = gBall.mGetOrientation().fPosition;
			gLookOffset -= gLookFrom;
			gLookOffset.mNormalize();
			gLookAt = gLookFrom + gLookOffset;

			gPitch = asin(gLookOffset.fZ);
			gYaw = acos(gLookOffset.fY /
				sqrt(gLookOffset.fX * gLookOffset.fX + gLookOffset.fY * gLookOffset.fY));
			if(gLookOffset.fX < 0.0)
				gYaw = 2.0 * kPI - gYaw;
		}
		else if(gHorizontalControl || gVerticalControl)
		{	
			gYaw += lDeltaTime * gHorizontalControl * kHorizontalSpeed;
			while(gYaw >= 2. * kPI) gYaw -= 2. * kPI;
			while(gYaw < 0) gYaw += 2. * kPI;
			gPitch += lDeltaTime * gVerticalControl * kVerticalSpeed;
			if(gPitch >= (kPI / 2.0)) gPitch = (kPI / 2.0) - 0.00001;
			if(gPitch <= -(kPI / 2.0)) gPitch = -(kPI / 2.0) + 0.00001;

			gLookOffset.fX = sin(gYaw);
			gLookOffset.fY = cos(gYaw);
			gLookOffset *= sqrt(cos(gPitch));
			gLookOffset.fZ = sin(gPitch);
			gLookOffset.fW = 0.0;
			gLookOffset.mNormalize();

			gLookAt = gLookFrom + gLookOffset;
		}
	}
}

void gInitSides()
{
	gSides[0] = kPoint0 - kVectorX * gFieldHalfWidth;
	gSides[1] = kPoint0 + kVectorX * gFieldHalfWidth;
	gSides[2] = kPoint0 - kVectorY * gFieldHalfWidth;
	gSides[3] = kPoint0 + kVectorY * gFieldHalfWidth;
	gSides[4] = kPoint0 - kVectorZ * gFieldHalfHeight;
	gSides[5] = kPoint0 + kVectorZ * gFieldHalfHeight;

	gSideNormals[0] = kVectorX;
	gSideNormals[1] = -kVectorX;
	gSideNormals[2] = kVectorY;
	gSideNormals[3] = -kVectorY;
	gSideNormals[4] = kVectorZ;
	gSideNormals[5] = -kVectorZ;

	gSideCorners[0] = gSides[0] + kVectorY * gFieldHalfWidth + kVectorZ * gFieldHalfHeight;
	gSideCorners[1] = gSides[1] - kVectorY * gFieldHalfWidth + kVectorZ * gFieldHalfHeight;
	gSideCorners[2] = gSides[2] + kVectorX * gFieldHalfWidth + kVectorZ * gFieldHalfHeight;
	gSideCorners[3] = gSides[3] - kVectorX * gFieldHalfWidth + kVectorZ * gFieldHalfHeight;
	gSideCorners[4] = gSides[4] + kVectorX * gFieldHalfWidth + kVectorY * gFieldHalfWidth;
	gSideCorners[5] = gSides[5] - kVectorX * gFieldHalfWidth + kVectorY * gFieldHalfWidth;

	gSideAdjacentCorners[0] = gSides[0] + kVectorY * gFieldHalfWidth - kVectorZ * gFieldHalfHeight;
	gSideAdjacentCorners[1] = gSides[1] - kVectorY * gFieldHalfWidth - kVectorZ * gFieldHalfHeight;
	gSideAdjacentCorners[2] = gSides[2] + kVectorX * gFieldHalfWidth - kVectorZ * gFieldHalfHeight;
	gSideAdjacentCorners[3] = gSides[3] - kVectorX * gFieldHalfWidth - kVectorZ * gFieldHalfHeight;
	gSideAdjacentCorners[4] = gSides[4] + kVectorX * gFieldHalfWidth - kVectorY * gFieldHalfWidth;
	gSideAdjacentCorners[5] = gSides[5] - kVectorX * gFieldHalfWidth - kVectorY * gFieldHalfWidth;

	for(int i = 0; i < 6; ++i)
	{
		gPlanes[i].mInit(gSides[i], gSideCorners[i], gSideAdjacentCorners[i]);
	}
}

void loop()
{
	static unsigned int i = 0;

	if(!gActive)
	{
		SDL_Quit();
		return;
	}

	SDL_Event lEvent;
	while(SDL_PollEvent(&lEvent))
	{
		switch(lEvent.type)
		{
			case SDL_QUIT:
				{
					gActive = false;
					SDL_Quit();
					return;
				}
				break;
			case SDL_KEYUP:
				{
					gProcessKeyUp(((SDL_KeyboardEvent *) &lEvent)->keysym.sym, 0, 0);
				}
				break;
			case SDL_KEYDOWN:
				{
					gProcessKeyDown(((SDL_KeyboardEvent *) &lEvent)->keysym.sym, 0, 0);
				}
				break;
			case SDL_VIDEORESIZE:
				{
					SDL_ResizeEvent &lResizeEvent = *((SDL_ResizeEvent *)(&lEvent));
					gReshape(lResizeEvent.w, lResizeEvent.h);
				}
				break;
			default:
				break;
		}
	}
	if(glGetError() != GL_NO_ERROR)
		fprintf(stderr, "\nGL Error: 0x%08x\n", glGetError());
	gSimulate();
	gRender();
	
	SDL_GL_SwapBuffers();
}

int main(int argc, char *argv[])
{
	if(SDL_Init(SDL_INIT_VIDEO) == -1)
	{
		fprintf(stderr, "SDL Error: %s", SDL_GetError());
			SDL_Quit();
			return 1;
	}

	if(SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1) == -1)
		fprintf(stderr, "SDL Error: %s", SDL_GetError());
	if(SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 ) == -1)
		fprintf(stderr, "SDL Error: %s", SDL_GetError());
	if(SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 5) == -1)
		fprintf(stderr, "SDL Error: %s", SDL_GetError());
	if(SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 ) == -1)
		fprintf(stderr, "SDL Error: %s", SDL_GetError());
	if(SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 ) == -1)
		fprintf(stderr, "SDL Error: %s", SDL_GetError());

	gpSurface = SDL_SetVideoMode(640, 480, 0, SDL_ANYFORMAT | SDL_OPENGL);
	if(!gpSurface)
	{
		fprintf(stderr, "SDL Error: %s", SDL_GetError());
		SDL_Quit();
		return 1;
	}

	gFragmentShader = gCreateShader(GL_FRAGMENT_SHADER, kFragmentShaderSource);
	gVertexShader = gCreateShader(GL_VERTEX_SHADER, kVertexShaderSource);
	gShaderProgram = gCreateProgram(gVertexShader, gFragmentShader);
	glValidateProgram(gShaderProgram);
	glUseProgram(gShaderProgram);	

	gReshape(640, 480);

	gGetGLMatrix(gObjectMat, kMatrix1);

	gInitSides();
	gDrawBall.mInit(10, 10, 1.0);

	gBall.mAddPart(Vector(0.0, 0.0, 0.0, 1.0));
	gBall.mSetMass(0.5 * kPI); // Radius is one.
	gBall.mSetInertia(kMatrix1 * 0.4 * gBall.mGetMass() * (1.0 * 1.0)); // (2/5) * M * R^2

	printf("Arrow keys rotate the camera.\n"
		"'a' moves forward. 's' moves backward.\n"
		"' ' pushes the ball in the direction your are facing.\n"
		"'p' toggles a pause in the physics simulation.\n"
		"'f' toggles the camera following the ball (rotation, not translation).\n"
		"'g' toggles gravity on and off.\n"
		"'j' jumps the camera to the balls current location (and looks where it's going if it's moving fast).\n"
		"'r' resets the simulation.\n");
		

#ifdef EMSCRIPTEN
	emscripten_set_main_loop(loop,60,true);
#else //EMSCRIPTEN
	while(true) loop();
#endif //EMSCRIPTEN

	// TODO: Clean shutdown.

	return 0;
}
